use std::{fs, rc::Rc, thread};

use crossbeam_channel::{bounded, Sender};
use libupgr::config::{Config, Step};
use miette::{miette, IntoDiagnostic, WrapErr};
use mlua::Lua;

pub fn start() -> miette::Result<LuaActorHandle> {
    let (lua_tx, lua_rx) = bounded(16);
    let (startup_tx, statup_rx) = bounded(0);
    thread::spawn(move || {
        let mut actor = match LuaActor::init() {
            Ok(a) => {
                startup_tx.send(Ok(())).unwrap();
                a
            },

            Err(e) => {
                startup_tx.send(Err(e)).unwrap();
                return;
            },
        };

        while let Ok(msg) = lua_rx.recv() {
            match actor.handle_msg(msg) {
                Err(e) => eprintln!("Error:{}", e),
                Ok(true) => break,
                Ok(false) => {},
            }
        }
    });

    statup_rx.recv().unwrap()?;

    Ok(LuaActorHandle { channel: lua_tx })
}

pub struct LuaActorHandle {
    channel: Sender<LuaMsg>,
}

impl LuaActorHandle {
    pub fn send(&self, msg: LuaMsg) {
        self.channel.send(msg).unwrap();
    }

    pub fn with_steps_sync<F, T>(&self, f: F) -> T
    where
        T: Send + 'static,
        for<'a> F: FnOnce(&'a mut Vec<Step>) -> T + Send + 'static,
    {
        let (tx, rx) = bounded(0);

        self.send(LuaMsg::WithSteps(Box::new(move |steps| {
            tx.send(f(steps)).unwrap()
        })));

        rx.recv().unwrap()
    }
}

struct LuaActor {
    _lua: &'static Lua,
    config: Config<'static>,
}

impl LuaActor {
    fn handle_msg(&mut self, msg: LuaMsg) -> miette::Result<bool> {
        match msg {
            LuaMsg::Exit => return Ok(true),
            LuaMsg::WithSteps(f) => f(&mut self.config.steps),
            LuaMsg::GetShell(tx) => tx.send(self.config.shell.clone()).unwrap(),
        }

        Ok(false)
    }

    fn init() -> miette::Result<Self> {
        let db = sled::open(
            dirs::cache_dir()
                .ok_or_else(|| miette!("Couldn't get cache dir"))?
                .join("upgr/db"),
        )
        .into_diagnostic()
        .wrap_err("Couldn't open database")?;

        let cfg_path = dirs::config_dir()
            .ok_or_else(|| miette!("Couldn't get config dir"))?
            .join("upgr/config.lua");

        let lua = unsafe { Lua::unsafe_new() };
        let lua = Box::leak(Box::new(lua));

        lua.globals()
            .set("upgr", libupgr::luautil::get_obj(lua, Rc::new(db))?)
            .into_diagnostic()?;

        let config = libupgr::config::run_config(
            lua,
            &fs::read(cfg_path)
                .into_diagnostic()
                .wrap_err("Couldn't read config")?,
        )
        .wrap_err("Failed to run config")?;

        Ok(LuaActor { _lua: lua, config })
    }
}

pub enum LuaMsg {
    Exit,
    WithSteps(Box<dyn FnOnce(&mut Vec<Step>) + Send>),
    GetShell(Sender<Vec<String>>),
}
