use relm4::{
    gtk::{self, prelude::*},
    ComponentUpdate,
    Model,
    Widgets,
};

use super::{AppModel, AppMsg};

pub struct EntryModel {
    id: Option<usize>,
    label: String,
    workdir: Option<String>,
    status: EntryStatus,
}

impl Model for EntryModel {
    type Msg = EntryMsg;
    type Widgets = EntryWidgets;
    type Components = ();
}

pub enum EntryMsg {
    SetInitialData {
        id: usize,
        label: String,
        workdir: Option<String>,
    },
    SetStatus(EntryStatus),
}

impl ComponentUpdate<AppModel> for EntryModel {
    fn init_model(_parent_model: &AppModel) -> Self {
        Self {
            id: None,
            label: String::new(),
            workdir: None,
            status: EntryStatus::Waiting,
        }
    }

    fn update(
        &mut self,
        msg: Self::Msg,
        _components: &Self::Components,
        _sender: relm4::Sender<Self::Msg>,
        _parent_sender: relm4::Sender<AppMsg>,
    ) {
        match msg {
            EntryMsg::SetInitialData { id, label, workdir } => {
                self.id = Some(id);
                self.label = label;
                self.workdir = workdir;
            },
            EntryMsg::SetStatus(status) => self.status = status,
        }
    }
}

#[relm4::widget(pub)]
impl Widgets<EntryModel, AppModel> for EntryWidgets {
    view! {
        gtk::Box {
            set_orientation: gtk::Orientation::Horizontal,

            append: stack = &gtk::Stack {
                set_transition_type: gtk::StackTransitionType::SlideLeftRight,

                add_child: waiting = &gtk::Image::from_icon_name("document-open-recent-symbolic") {},
                add_child: running = &gtk::Spinner { set_spinning: true },
                add_child: success = &gtk::Image::from_icon_name("object-select") {},
                add_child: error = &gtk::Image::from_icon_name("dialog-error") {},
                add_child: skipped = &gtk::Image::from_icon_name("media-skip-forward") {},
            },

            append = &gtk::Box {
                set_orientation: gtk::Orientation::Vertical,

                append = &gtk::Label {
                    set_halign: gtk::Align::Start,
                    set_label: watch! { &model.label },
                },

                append = &gtk::Label {
                    set_halign: gtk::Align::Start,
                    add_css_class: "dim-label",
                    set_label: watch! {
                        &model
                            .workdir
                            .as_ref()
                            .cloned()
                            .unwrap_or_else(|| {
                                std::env::current_dir()
                                    .map(|p| p.to_string_lossy().to_string())
                                    .unwrap_or_else(|_| String::new(),
                                )
                            })
                    },
                }
            }
        }
    }

    fn post_view() {
        match model.status {
            EntryStatus::Waiting => self.stack.set_visible_child(&self.waiting),
            EntryStatus::Running => self.stack.set_visible_child(&self.running),
            EntryStatus::Success => self.stack.set_visible_child(&self.success),
            EntryStatus::Error => self.stack.set_visible_child(&self.error),
            EntryStatus::Skipped => self.stack.set_visible_child(&self.skipped),
        }
    }
}

pub enum EntryStatus {
    Waiting,
    Running,
    Success,
    Error,
    Skipped,
}
