use std::cell::Cell;

use crossbeam_channel::bounded;
use libupgr::config::CfgCommand;
use relm4::{
    gtk::{self, prelude::*},
    send,
    AppUpdate,
    Components,
    Model,
    RelmComponent,
    Widgets,
};

use crate::lua_actor::{LuaActorHandle, LuaMsg};

use self::{
    entry::{EntryModel, EntryMsg, EntryStatus},
    terminal::{SpawnData, TerminalModel, TerminalMsg},
};

mod entry;
mod terminal;

pub struct AppModel {
    pub can_run: bool,
    pub lua_actor: LuaActorHandle,
    pub selected_idx: i32,
    pub idx_changed: Cell<IdxChanged>,
}

impl Model for AppModel {
    type Msg = AppMsg;
    type Widgets = AppWidgets;
    type Components = AppComponents;
}

impl AppUpdate for AppModel {
    fn update(
        &mut self,
        msg: Self::Msg,
        components: &Self::Components,
        sender: relm4::Sender<Self::Msg>,
    ) -> bool {
        match msg {
            AppMsg::TerminalExited { id, success } => {
                if let Some(e) = components.entries.get(id) {
                    e.send(EntryMsg::SetStatus(if success {
                        EntryStatus::Success
                    } else {
                        EntryStatus::Error
                    }))
                    .unwrap();
                }

                if id + 1 >= components.terminals.len() {
                    self.can_run = true;
                } else {
                    send!(sender, AppMsg::RunId(id + 1));
                }
            },

            AppMsg::RunId(id) => {
                self.can_run = false;
                if id == 0 {
                    for e in &components.entries {
                        e.send(EntryMsg::SetStatus(EntryStatus::Waiting)).unwrap();
                    }
                }

                components.entries[id]
                    .send(EntryMsg::SetStatus(EntryStatus::Running))
                    .unwrap();

                send!(sender, AppMsg::RowSelected(id as i32, IdxChanged::Auto));

                self.lua_actor
                    .send(LuaMsg::WithSteps(Box::new(move |steps| {
                        if let Some(when) = &steps[id].when {
                            if when.call(()).unwrap_or(false) {
                                send!(sender, AppMsg::ExecId(id));
                            } else {
                                send!(sender, AppMsg::SkipId(id));
                            }
                        } else {
                            send!(sender, AppMsg::ExecId(id));
                        }
                    })));
            },

            AppMsg::SkipId(id) => {
                components.entries[id]
                    .send(EntryMsg::SetStatus(EntryStatus::Skipped))
                    .unwrap();
                if id + 1 >= components.terminals.len() {
                    self.can_run = true;
                } else {
                    send!(sender, AppMsg::RunId(id + 1));
                }
            },

            AppMsg::ExecId(id) => {
                components.entries[id]
                    .send(EntryMsg::SetStatus(EntryStatus::Running))
                    .unwrap();

                let (tx, rx) = bounded(0);
                self.lua_actor.send(LuaMsg::GetShell(tx));
                let mut shell = rx.recv().unwrap();

                let (argv, workdir) = self.lua_actor.with_steps_sync(move |steps| {
                    let step = &steps[id];

                    let argv = match step.command.clone() {
                        CfgCommand::Shell(c) => {
                            shell.push(c);
                            shell
                        },
                        CfgCommand::Args(a) => a,
                    };

                    (argv, step.workdir.clone().map(Into::into))
                });

                components.terminals[id]
                    .send(TerminalMsg::Spawn(SpawnData { argv, workdir }))
                    .unwrap();
            },
            AppMsg::RowSelected(idx, idxc) => {
                self.selected_idx = idx;
                self.idx_changed.set(idxc);
            },
        }
        true
    }
}

#[relm4::widget(pub)]
impl Widgets<AppModel, ()> for AppWidgets {
    view! {
        main_window = gtk::ApplicationWindow {
            set_title: Some("gupgr"),

            set_child = Some(&gtk::Box) {
                set_orientation: gtk::Orientation::Horizontal,

                append = &gtk::Box  {
                    set_orientation: gtk::Orientation::Vertical,

                    append = &gtk::ScrolledWindow {
                        set_vexpand: true,
                        set_width_request: 500,

                        set_child: list_box = Some(&gtk::ListBox) {
                            append: iterate!(
                                components
                                    .entries
                                    .iter()
                                    .map(RelmComponent::root_widget)
                            ),

                            connect_row_selected(sender) => move |_, row| {
                                if let Some(row) = row {
                                    send!(sender, AppMsg::RowSelected(row.index(), IdxChanged::Manual));
                                }
                            },
                        }
                    },

                    append = &gtk::Button {
                        set_label: "Run!",

                        set_sensitive: watch!(model.can_run),

                        connect_clicked(sender) => move |_| send!(sender, AppMsg::RunId(0)),
                    }
                },

                append: stack = &gtk::Stack {
                    set_transition_type: gtk::StackTransitionType::SlideUpDown,
                    add_child: iterate!(
                        components
                            .terminals
                            .iter()
                            .map(|t| {
                                let widget = t.root_widget();
                                stack_entries.push(widget.clone());
                                widget
                            })
                    ),
                }
            }
        }
    }

    additional_fields! {
        stack_entries: Vec<vte4::Terminal>,
    }

    fn pre_init() {
        let mut stack_entries = vec![];
    }

    fn post_view() {
        match model.idx_changed.take() {
            IdxChanged::No => {},
            IdxChanged::Manual => {
                self.stack
                    .set_visible_child(&self.stack_entries[model.selected_idx as usize]);
            },
            IdxChanged::Auto => self
                .list_box
                .select_row(self.list_box.row_at_index(model.selected_idx).as_ref()),
        }
    }
}

pub enum AppMsg {
    TerminalExited { id: usize, success: bool },
    RunId(usize),
    SkipId(usize),
    ExecId(usize),
    RowSelected(i32, IdxChanged),
}

pub struct AppComponents {
    entries: Vec<RelmComponent<EntryModel, AppModel>>,
    terminals: Vec<RelmComponent<TerminalModel, AppModel>>,
}

impl Components<AppModel> for AppComponents {
    fn init_components(parent_model: &AppModel, parent_sender: relm4::Sender<AppMsg>) -> Self {
        let msgs = parent_model.lua_actor.with_steps_sync(|steps| {
            steps
                .iter()
                .enumerate()
                .map(|(id, step)| {
                    (
                        EntryMsg::SetInitialData {
                            id,
                            label: step.command.to_string(),
                            workdir: step
                                .workdir
                                .as_ref()
                                .map(|d| d.to_string_lossy().to_string()),
                        },
                        TerminalMsg::SetId(id),
                    )
                })
                .collect::<Vec<_>>()
        });

        let mut entries = Vec::with_capacity(msgs.len());
        let mut terminals = Vec::with_capacity(msgs.len());

        for (e_msg, t_msg) in msgs {
            let entr = RelmComponent::new(parent_model, parent_sender.clone());
            let term = RelmComponent::new(parent_model, parent_sender.clone());

            entr.send(e_msg).unwrap();
            term.send(t_msg).unwrap();

            entries.push(entr);
            terminals.push(term);
        }

        Self { entries, terminals }
    }

    fn connect_parent(&mut self, parent_widgets: &AppWidgets) {
        for entry in &mut self.entries {
            entry.connect_parent(parent_widgets)
        }

        for terminal in &mut self.terminals {
            terminal.connect_parent(parent_widgets)
        }
    }
}

pub enum IdxChanged {
    No,
    Auto,
    Manual,
}

impl Default for IdxChanged {
    fn default() -> Self {
        Self::No
    }
}
