use std::{cell::RefCell, ffi::OsString, os::unix::ffi::OsStrExt, path::Path};

use relm4::{
    gtk::{self, glib::SpawnFlags, prelude::*},
    send,
    ComponentUpdate,
    Model,
    Widgets,
};
use vte4::{PtyFlags, TerminalExt};

use super::{AppModel, AppMsg};

pub struct TerminalModel {
    id: Option<usize>,
    do_spawn: RefCell<Option<SpawnData>>,
}

impl Model for TerminalModel {
    type Msg = TerminalMsg;
    type Widgets = TerminalWidgets;
    type Components = ();
}

pub enum TerminalMsg {
    SetId(usize),
    Exit(i32),
    Spawn(SpawnData),
}

pub struct SpawnData {
    pub argv: Vec<String>,
    pub workdir: Option<OsString>,
}

impl ComponentUpdate<AppModel> for TerminalModel {
    fn init_model(_parent_model: &AppModel) -> Self {
        Self {
            id: None,
            do_spawn: RefCell::new(None),
        }
    }

    fn update(
        &mut self,
        msg: Self::Msg,
        _components: &Self::Components,
        _sender: relm4::Sender<Self::Msg>,
        parent_sender: relm4::Sender<AppMsg>,
    ) {
        match msg {
            TerminalMsg::SetId(id) => self.id = Some(id),
            TerminalMsg::Exit(status) => {
                send!(
                    parent_sender,
                    AppMsg::TerminalExited {
                        id: self.id.unwrap(),
                        success: status == 0,
                    }
                );
            },
            TerminalMsg::Spawn(s) => *self.do_spawn.borrow_mut() = Some(s),
        }
    }
}

#[relm4::widget(pub)]
impl Widgets<TerminalModel, AppModel> for TerminalWidgets {
    view! {
        term = vte4::Terminal {
            set_hexpand: true,
            set_vexpand: true,

            connect_child_exited(sender) => move |_, status| {
                send!(sender, TerminalMsg::Exit(status));
            },
        }
    }

    fn post_view() {
        if let Some(SpawnData { argv, workdir }) = model.do_spawn.take() {
            let argv = argv.iter().map(Path::new).collect::<Vec<_>>();
            let envv_data = std::env::vars_os()
                .map(|(mut a, b)| {
                    a.push("=");
                    a.push(b);
                    a
                })
                .collect::<Vec<_>>();
            let envv = envv_data.iter().map(Path::new).collect::<Vec<_>>();

            self.term.spawn_async(
                PtyFlags::empty(),
                workdir
                    .as_ref()
                    .map(|s| unsafe { std::str::from_utf8_unchecked(s.as_bytes()) }),
                &argv,
                &envv,
                SpawnFlags::DO_NOT_REAP_CHILD, // needed with child_exited is connected
                Some(Box::new(|| {})),
                -1,
                gtk::gio::Cancellable::NONE,
                None,
            );
        }
    }
}
