use crate::output::{info_table::InfoTable, FormattedCfgCommand};
use clap::Parser;
use libupgr::{
    config::{CfgCommand, Config},
    luautil,
};
use miette::{miette, Context, IntoDiagnostic};
use mlua::{prelude::LuaFunction, Lua};
use owo_colors::OwoColorize;
use std::{
    fs::File,
    io::{self, Read},
    path::PathBuf,
    rc::Rc,
    time::{Duration, Instant},
};
use terminal_size::terminal_size;

mod output;

#[derive(Parser)]
struct Opt {
    #[clap(long, short)]
    /// Don't run interactive steps, unless they have unint_alt set
    uninteractive: bool,

    #[clap(long, short)]
    /// The config file to use. Defaults to ~/.config/upgr/config.lua
    config: Option<PathBuf>,
}

fn main() -> miette::Result<()> {
    let opt = Opt::parse();

    let cfg_path = if let Some(path) = opt.config {
        path
    } else {
        dirs::config_dir()
            .ok_or_else(|| miette!("Couldn't get config directory"))?
            .join("upgr/config.lua")
    };

    let mut file = File::open(cfg_path)
        .into_diagnostic()
        .wrap_err("Couldn't open config")?;

    let mut config_bytes = vec![];
    file.read_to_end(&mut config_bytes)
        .into_diagnostic()
        .wrap_err("Failed to read config")?;

    let db = Rc::new(
        sled::open(
            dirs::cache_dir()
                .ok_or_else(|| miette!("Couldn't get cache dir"))?
                .join("upgr/db"),
        )
        .into_diagnostic()
        .wrap_err("Couldn't open database")?,
    );

    let lua = unsafe { Lua::unsafe_new() };
    lua.globals()
        .set(
            "upgr",
            luautil::get_obj(&lua, Rc::clone(&db)).wrap_err("Couldn't create lua upgr object")?,
        )
        .into_diagnostic()
        .wrap_err("Couldn't set upgr lua object")?;

    let config =
        libupgr::config::run_config(&lua, &config_bytes).wrap_err("Failed to run config")?;
    exec_steps(config, opt.uninteractive).wrap_err("Failed to execute steps")?;
    db.flush()
        .into_diagnostic()
        .wrap_err("Failed to flush DB")?;

    Ok(())
}

pub fn exec_steps(conf: Config, uninteractive: bool) -> miette::Result<()> {
    for step in conf.steps {
        if step.interactive && uninteractive {
            if let Some(cmd) = step.unint_alt {
                exec_cmd(cmd, &conf.shell, step.when, step.workdir)?;
            }
        } else {
            exec_cmd(step.command, &conf.shell, step.when, step.workdir)?;
        }
    }

    Ok(())
}

pub fn exec_cmd(
    cfg_cmd: CfgCommand,
    shell: &[String],
    when: Option<LuaFunction>,
    workdir: Option<PathBuf>,
) -> miette::Result<()> {
    if let Some(func) = when {
        let val: bool = func
            .call(())
            .into_diagnostic()
            .wrap_err("Failed to call when function")?;

        if !val {
            println!(
                "Skipping {}",
                FormattedCfgCommand::new(&cfg_cmd, Some(shell))
            );
            return Ok(());
        }
    }

    let mut table = InfoTable::with_title("Running");
    table.row("Command", FormattedCfgCommand::new(&cfg_cmd, Some(shell)));

    let mut cmd = cfg_cmd.as_command(shell)?;
    if let Some(workdir) = workdir {
        table.row("Workdir", &workdir.to_string_lossy().as_ref());
        cmd.current_dir(workdir);
    }

    let (terminal_size::Width(width), _) =
        terminal_size().ok_or_else(|| miette!("Couldn't get terminal size"))?;

    table
        .write_to(io::stdout(), width as usize)
        .into_diagnostic()?;

    let start_time = Instant::now();
    let exit = cmd.spawn().into_diagnostic()?.wait().into_diagnostic()?;
    let total_time = Instant::now() - start_time;

    InfoTable::with_title("Summary")
        .row("Command", FormattedCfgCommand::new(&cfg_cmd, Some(shell)))
        .row(
            "Time",
            humantime::format_duration(
                // only be precise to milliseconds
                Duration::from_millis(total_time.as_millis() as u64),
            )
            .blue(),
        )
        .row(
            "Exit Code",
            &exit
                .code()
                .map(|s| {
                    if s == 0 {
                        0.green().to_string()
                    } else {
                        s.red().to_string()
                    }
                })
                .unwrap_or_else(|| "Unknown".blue().to_string()),
        )
        .write_to(io::stdout(), width as usize)
        .into_diagnostic()?;

    Ok(())
}
