use std::{
    cell::Cell,
    fmt::{Display, Write},
};

use libupgr::config::CfgCommand;
use owo_colors::OwoColorize;

pub mod info_table;

pub struct FormattedCfgCommand<'a, Shell> {
    command: &'a CfgCommand,
    shell: Cell<Option<Shell>>,
}

impl<'a, Shell, S> FormattedCfgCommand<'a, Shell>
where
    Shell: IntoIterator<Item = S>,
    S: Display,
{
    pub fn new(command: &'a CfgCommand, shell: Option<Shell>) -> Self {
        Self {
            command,
            shell: Cell::new(shell),
        }
    }
}

impl<'a, Shell, S> Display for FormattedCfgCommand<'a, Shell>
where
    Shell: IntoIterator<Item = S>,
    S: Display,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self.command {
            CfgCommand::Args(args) => {
                FormattedEscapedCommand::new(args).cyan().fmt(f)?;
            },
            CfgCommand::Shell(cmd) => {
                if let Some(shell) = self.shell.take() {
                    write!(f, "[{}] ", FormattedEscapedCommand::new(shell).red())?;
                }

                cmd.cyan().fmt(f)?;
            },
        }

        Ok(())
    }
}

struct FormattedEscapedCommand<Shell>(Cell<Option<Shell>>);

impl<Shell> FormattedEscapedCommand<Shell> {
    fn new(s: Shell) -> Self {
        Self(Cell::new(Some(s)))
    }
}

impl<Shell, S> Display for FormattedEscapedCommand<Shell>
where
    Shell: IntoIterator<Item = S>,
    S: Display,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut first = true;
        for a in self
            .0
            .take()
            .expect("Attempt to format FormattedEscapedCommand twice")
        {
            let a = a.to_string();
            if !first {
                f.write_char(' ')?;
            }
            first = false;

            if a.contains(' ') {
                write!(f, r#""{}""#, a)?;
            } else {
                f.write_str(&a)?;
            }
        }
        Ok(())
    }
}
